@@ -4,13 +4,14 @@ import numpy as np
 from gymnasium import spaces
 from ray.rllib import env
 import collections
+from ray.rllib.env import env_context
 
 from . import encoder, typing
 import os
 import platform
-import socket
 import subprocess
 import zipfile
+import portpicker
 
 from .game import constants, footsies_game
 from ..binary_manager import get_binary_manager
@@ -21,7 +22,6 @@ class FootsiesEnv(env.MultiAgentEnv):
     LINUX_ZIP_PATH_HEADLESS = "binaries/footsies_linux_server_021725.zip"
     LINUX_ZIP_PATH_WINDOWED = "binaries/footsies_linux_windowed_021725.zip"
     SPECIAL_CHARGE_FRAMES = 60
-    GUARD_BREAK_REWARD = 0.3
 
     observation_space = spaces.Dict(
         {
@@ -62,17 +62,27 @@ class FootsiesEnv(env.MultiAgentEnv):
         if config is None:
             config = {}
         self.config = config
+        self.return_fight_state_in_infos = config.get("return_fight_state_in_infos", False)
         self.use_build_encoding = config.get("use_build_encoding", False)
         self.agents: list[typing.AgentID] = ["p1", "p2"]
         self.possible_agents: list[typing.AgentID] = self.agents.copy()
         self._agent_ids: set[typing.AgentID] = set(self.agents)
+        self.win_reward_scaling_coeff = self.config.get("win_reward_scaling_coeff", 10.0)
+        self.guard_break_reward_value = self.config.get("guard_break_reward", 0)
+        self.use_reward_budget = self.config.get("use_reward_budget", True)
+        assert self.guard_break_reward_value * 3 < self.win_reward_scaling_coeff, (
+            "Guard break reward total must be less than the win reward (guard break reward * 3 < win reward)"
+        )
+
+        self.reward_budget = {agent: self.win_reward_scaling_coeff for agent in self.agents}
+
 
         self.evaluation = config.get("evaluation", False)
 
         self.t: int = 0
         self.max_t: int = config.get("max_t", 1000)
         self.frame_skip: int = config.get("frame_skip", 4)
-        self.action_delay_frames: int = config.get("action_delay", 16)
+        self.action_delay_frames: int = config.get("action_delay", 8)
 
         assert (
             self.action_delay_frames % self.frame_skip == 0
@@ -81,23 +91,15 @@ class FootsiesEnv(env.MultiAgentEnv):
         self.action_delay_steps: int = self.action_delay_frames // self.frame_skip
         self.encoder = encoder.FootsiesEncoder()
         self._action_queues: dict[typing.AgentID, collections.deque[int]] = None
+        self.prev_actions: dict[typing.AgentID, int] = {agent: constants.EnvActions.NONE for agent in self.agents}
         self._reset_action_delay_queues()
 
 
-        
         port = config.get("port", None)
         self.headless = config.get("headless", True)
-        # we'll start game servers at 50051 for training
-        # and 40051 for evaluation. Worker index starts at
-        # 1, so we won't see 40050/50050.
+        # Use portpicker to automatically find an available port
         if port is None:
-            start_port = 40050 if self.evaluation else 50050
-            port = (
-                start_port
-                + int(config.get("worker_index", 0))
-                * config.get("num_envs_per_worker", 1)
-                + config.get("vector_index", 0)
-            )
+            port = portpicker.pick_unused_port()
 
         # If specified, we'll launch the binaries from the environment itself.
         self.server_process = None
@@ -111,19 +113,45 @@ class FootsiesEnv(env.MultiAgentEnv):
         )
 
         self.last_game_state = None
-        self.special_charge_queue = {
-            "p1": -1,
-            "p2": -1,
+        self._holding_special_charge = {
+            "p1": False,
+            "p2": False,
         }
 
-    def _is_port_in_use(self, port: int) -> bool:
-        """Check if a port is already in use."""
-        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
-            try:
-                sock.bind(('localhost', port))
-                return False
-            except OSError:
-                return True
+
+    def _get_fight_state_dicts(self):
+        """
+        class FightState:
+            distance_x: float
+            is_opponent_damage: bool 
+            is_opponent_guard_break: bool
+            is_opponent_blocking: bool
+            is_opponent_normal_attack: bool
+            is_opponent_special_attack: bool
+            is_facing_right: bool
+        """
+        fight_state_dict = {
+            "p1": {},
+            "p2": {},
+        }
+        p1_state, p2_state = self.last_game_state.player1, self.last_game_state.player2
+
+        dist_x = np.abs(p1_state.player_position_x - p2_state.player_position_x)
+
+        for player, opp_state in zip(["p1", "p2"], [p2_state, p1_state]):
+            fight_state_dict[player]["distance_x"] = dist_x
+            fight_state_dict[player]["is_opponent_damage"] = opp_state.current_action_id == constants.ActionID.DAMAGE
+            fight_state_dict[player]["is_opponent_guard_break"] = opp_state.current_action_id == constants.ActionID.GUARD_BREAK
+            fight_state_dict[player]["is_opponent_blocking"] = opp_state.current_action_id in [constants.ActionID.GUARD_CROUCH, constants.ActionID.GUARD_STAND, constants.ActionID.GUARD_M]
+            fight_state_dict[player]["is_opponent_normal_attack"] = opp_state.current_action_id in [constants.ActionID.N_ATTACK, constants.ActionID.B_ATTACK]
+            fight_state_dict[player]["is_opponent_special_attack"] = opp_state.current_action_id in [constants.ActionID.N_SPECIAL, constants.ActionID.B_SPECIAL]
+    
+
+        for player, state in zip(["p1", "p2"], [p1_state, p2_state]):
+            fight_state_dict[player]["is_facing_right"] = state.is_face_right
+
+
+        return fight_state_dict
 
     def _launch_binaries(self, port: int):
         # Check if we're on a supported platform
@@ -162,10 +190,7 @@ class FootsiesEnv(env.MultiAgentEnv):
             # If not, make it executable
             os.chmod(binary_path, 0o755)
 
-        # Check if the port is already in use
-        if self._is_port_in_use(port):
-            print(f"Port {port} is already in use. Skipping binary launch.")
-            return
+        # portpicker already ensures the port is available, so no need to check
 
         command = [binary_path, "--port", str(port)]
         
@@ -233,7 +258,7 @@ class FootsiesEnv(env.MultiAgentEnv):
                 " Observed {len(self._action_queues[agent_id])}, expected {self.action_delay_steps}"
             )
 
-    def get_obs(self, game_state):
+    def get_obs(self, game_state, prev_actions, is_charging_special: dict[typing.AgentID, bool]):
         if self.use_build_encoding:
             raise NotImplementedError(
                 "Build encoder has not yet integrated action delay! "
@@ -250,7 +275,7 @@ class FootsiesEnv(env.MultiAgentEnv):
             # }
             # return encoded_state_dict
         else:
-            return self.encoder.encode(game_state)
+            return self.encoder.encode(game_state, prev_actions, is_charging_special)
 
     def reset(
         self,
@@ -272,14 +297,16 @@ class FootsiesEnv(env.MultiAgentEnv):
 
         self._reset_action_delay_queues()
 
+        # Reset reward budget
+        self.reward_budget = {agent: self.win_reward_scaling_coeff for agent in self.agents}
+
         self.encoder.reset()
 
-        if not self.use_build_encoding:
-            self.last_game_state = self.game.get_state()
+        self.last_game_state = self.game.get_state()
 
-        observations = self.get_obs(self.last_game_state)
+        observations = self.get_obs(self.last_game_state, self.prev_actions, self._holding_special_charge)
 
-        return observations, {agent: {} for agent in self.agents}
+        return observations, self.get_infos()
 
     def step(self, actions: dict[typing.AgentID, typing.ActionType]) -> tuple[
         dict[typing.AgentID, typing.ObsType],
@@ -297,7 +324,6 @@ class FootsiesEnv(env.MultiAgentEnv):
         """
         self.t += 1
 
-
         # Update action queue -> dequeue old actions, enqueue new actions
         actions_to_execute: dict[typing.AgentID, typing.ActionType] = {}
         if self.action_delay_frames == 0:
@@ -308,19 +334,19 @@ class FootsiesEnv(env.MultiAgentEnv):
                 self._action_queues[agent_id].append(actions[agent_id])
 
         for agent_id in self.agents:
-            empty_queue = self.special_charge_queue[agent_id] < 0
+            holding_special_charge = self._holding_special_charge[agent_id]
             action_is_special_charge = (
                 actions_to_execute[agent_id] == constants.EnvActions.SPECIAL_CHARGE
             )
 
-            # Refill the charge queue only if we're not already in a special charge.
-            if action_is_special_charge and empty_queue:
-                self.special_charge_queue[agent_id] = (
-                    self._build_charged_special_queue()
-                )
+            # Toggle the special charge based on whether or not we're holding special already
+            if action_is_special_charge and not holding_special_charge:
+                self._holding_special_charge[agent_id] = True
+            elif action_is_special_charge and holding_special_charge:
+                self._holding_special_charge[agent_id] = False
+                actions_to_execute[agent_id] = self.prev_actions[agent_id]
 
-            if self.special_charge_queue[agent_id] >= 0:
-                self.special_charge_queue[agent_id] -= 1
+            if self._holding_special_charge[agent_id]:
                 actions_to_execute[agent_id] = self._convert_to_charge_action(
                     actions_to_execute[agent_id]
                 )
@@ -331,30 +357,46 @@ class FootsiesEnv(env.MultiAgentEnv):
         game_state = self.game.step_n_frames(
             p1_action=p1_action, p2_action=p2_action, n_frames=self.frame_skip
         )
-        observations = self.get_obs(game_state)
+        observations = self.get_obs(game_state, actions_to_execute, self._holding_special_charge)
 
         terminated = game_state.player1.is_dead or game_state.player2.is_dead
 
-        # Zero-sum game: 1 if other player is dead, -1 if you're dead:
-        rewards = {
-            "p1": int(game_state.player2.is_dead)
-            - int(game_state.player1.is_dead),
-            "p2": int(game_state.player1.is_dead)
-            - int(game_state.player2.is_dead),
-        }
 
-        if self.config.get("reward_guard_break", False):
+        rewards = {a_id: 0.0 for a_id in self.agents} 
+        # Apply guard-break reward, if using. 
+        if self.guard_break_reward_value != 0:
             p1_prev_guard_health = self.last_game_state.player1.guard_health
             p2_prev_guard_health = self.last_game_state.player2.guard_health
             p1_guard_health = game_state.player1.guard_health
             p2_guard_health = game_state.player2.guard_health
 
+            # Guard break reward is deducted from the overall "budget" of reward
+            # to avoid biasing gameplay towards guard break. The total reward
+            # always remains the same, but we can make the signal more dense by 
+            # providing guard break rewards. This can be turned off with 
+            # "use_reward_budget=False" in the environment config.
             if p2_guard_health < p2_prev_guard_health:
-                rewards["p1"] += self.GUARD_BREAK_REWARD
-                rewards["p2"] -= self.GUARD_BREAK_REWARD
+                if self.use_reward_budget:
+                    self.reward_budget["p1"] -= self.guard_break_reward_value
+                rewards["p1"] += self.guard_break_reward_value
+                rewards["p2"] -= self.guard_break_reward_value
             if p1_guard_health < p1_prev_guard_health:
-                rewards["p2"] += self.GUARD_BREAK_REWARD
-                rewards["p1"] -= self.GUARD_BREAK_REWARD
+                if self.use_reward_budget:
+                    self.reward_budget["p2"] -= self.guard_break_reward_value
+                rewards["p2"] += self.guard_break_reward_value
+                rewards["p1"] -= self.guard_break_reward_value
+
+        # If the other player is dead, reward the player who is alive.
+        # We apply rewards as remaining_reward_budget * is_dead + guard_break. 
+        is_dead = {
+            "p1": int(game_state.player2.is_dead)
+            - int(game_state.player1.is_dead),
+            "p2": int(game_state.player1.is_dead)
+            - int(game_state.player2.is_dead),
+        }
+        rewards = {a_id: v + self.reward_budget[a_id] * is_dead[a_id] for a_id, v in rewards.items()}
+
+
 
         terminateds = {
             "p1": terminated,
@@ -370,6 +412,7 @@ class FootsiesEnv(env.MultiAgentEnv):
         }
 
         self.last_game_state = game_state
+        self.prev_actions = actions_to_execute
 
         # ~~~ For debugging game build! ~~~
         # encoded_state = self.game.get_encoded_state()
@@ -391,7 +434,10 @@ class FootsiesEnv(env.MultiAgentEnv):
         return observations, rewards, terminateds, truncateds, self.get_infos()
 
     def get_infos(self):
-        return {agent: {} for agent in self.agents}
+        infos = {agent: {} for agent in self.agents}
+        if self.return_fight_state_in_infos:
+            infos.update(self._get_fight_state_dicts())
+        return infos
 
     def _build_charged_special_queue(self):
         assert self.SPECIAL_CHARGE_FRAMES % self.frame_skip == 0
