@@ -3,3 +3,5 @@ from pettingzoo.utils import env
 AgentID = env.AgentID
 ObsType = env.ObsType
 ActionType = env.ActionType
+ActionBits = int
+EnvID = str | int
\ No newline at end of file
